/* Phrase.c -- Copyright 1989 Liam R. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

/*
 * Deal with (WID, FID, Offfset) triples
 * Liam Quin, September 1989
 *
 * $Id: Phrase.c,v 1.20 93/11/28 15:13:25 lee Exp $
 *
 */

#include "error.h"

#include <stdio.h> /* stderr, also for fileinfo.h */
#include <fcntl.h>
#include <malloc.h>
#include <sys/types.h>
#include <ctype.h>
#ifdef BSD
# include <strings.h>
#else
# include <string.h>
#endif
#include "globals.h" /* defines and declarations for database filenames */

#include "fileinfo.h" /* for wordinfo.h */
#include "wordinfo.h"
#include "pblock.h"
#include "phrase.h"
#include "wordrules.h"
#include "emalloc.h"

/** Unix system calls that need to be declared: **/
extern void exit();
/** Unix/C Library Functions: **/
extern unsigned int sleep();
#ifndef tolower
 extern int tolower();
#endif

/** lqtext functions: **/
extern API int TooCommon();
extern char *UnFlag();

/** functions within this file that need forward declarations **/
PRIVATE int CheckFlags();
/** **/


#ifndef MAXPHRASELEN
# define MAXPHRASELEN 2000
#endif

extern int AsciiTrace;

API t_Phrase *
String2Phrase(String)
    char *String;
{
    extern t_WordInfo *WID2WordInfo();
    extern t_WID Word2WID();
    extern t_WordInfo *SpReadWord();

    t_Phrase *Result;
    t_PhraseItem **ThisWord;
    /* (* 3 because in the worst case, "a a a" expands to "[a] [a] [a]") */
    char *q; /* not register, we take its address */
    char *Start = 0;
    int InWord = 0;

    Result = (t_Phrase *) emalloc("String2Phrase result", sizeof(t_Phrase));
    Result->Next = (t_Phrase *) 0;
    Result->HasUnknownWords = 0;

    *(ThisWord = &Result->Words) = (t_PhraseItem *) 0;

    /* initialise ReadWord(): */
    (void) SpReadWord((char *) 0, (char *) 0, 0);

    /* March along the supplied phrase, looking for keywords.
     * surround unindexed or short words with [brackets].
     * Also converts to lower case and strips plurals.
     */
    for (q = String; /*LOTSOFTIMES*/; ) {
	int UsedABracket = 0;
	t_WordInfo *WordInfo;

	WordInfo = SpReadWord(&q, &Start, 1); /* 1 for ignore common words */
	if (!WordInfo) {
	    break;
	}

	WordInfo->WID = Word2WID(WordInfo->Word, WordInfo->Length);
	if (WordInfo->WID > 0) {
	    t_WordInfo *W = WID2WordInfo(WordInfo->WID);
	    
	    if (W == (t_WordInfo *) 0) {
#ifdef ASCIITRACE
		if (AsciiTrace > 10) {
		    fprintf(stderr, " Unknown(%s) ", WordInfo->Word);
		}
#endif
		Result->HasUnknownWords++;
	    } else {
		int Flags = W->Flags;
		*ThisWord = (t_PhraseItem *)
				emalloc("Phrase Item", sizeof(t_PhraseItem));
		W->WordPlace = WordInfo->WordPlace; /* struct copy */
		W->WordPlace.Flags |= Flags;
		if (AsciiTrace > 1) {
		    fprintf(stderr, "Word %s --> %s, %lu matches\n",
			WordInfo->Word,
			UnFlag(W, W->WordPlace.Flags),
			W->NumberOfWordPlaces
		    );
#ifdef ASCIITRACE
		    fprintWordInfo(stderr, W, "String2Phrase");
#endif
		}
		/* point to the new space */
		(*ThisWord)->Word = W;
		(*ThisWord)->WordStart = Start;
		(*ThisWord)->Next = (t_PhraseItem *) 0;
		(*ThisWord)->SearchIndex = 0L;
		ThisWord = &(*ThisWord)->Next;

	    }
	} else {
	    Result->HasUnknownWords++;
	}
	if (!*q) break;
    } /* for */

    if (ThisWord == &Result->Words) {
	/* There were no words in the phrase! */
	return (t_Phrase *) 0; /* TODO plug momory leaks NOTDONE FIXME */
    }

    Result->OriginalString = emalloc("PhraseOriginalString",strlen(String) + 1);
    (void) strcpy(Result->OriginalString, String);
    Result->ModifiedString = emalloc("PhraseModifiedString",strlen(String) + 1);
    (void) strcpy(Result->ModifiedString, String);

    Result->NumberOfMatches = 0;
    Result->Matches = (t_MatchList *) 0;
    if (AsciiTrace > 1) {
	fprintf(stderr, "phrase \"%s\",\n", Result->OriginalString, String);
    }
    return Result;
}

API t_Answer *
GetFiles(Phrase)
    t_Phrase *Phrase;
{
    PRIVATE char *MakeOneDescription();

    t_Answer *Result = 0;
    t_Answer **RP = &Result;
    t_MatchList *MP;
    t_FID LastFID;
    unsigned long ThisFIDNumberOfMatches = 0L;

    if (!Phrase || !Phrase->Matches) {
	return Result;
    }

    LastFID = Phrase->Matches->Match->Where->FID;

    for (MP = Phrase->Matches; MP; MP = MP->Next) {
	if (MP->Match->Where->FID != LastFID) {
	    char *p;
	    
	    p = MakeOneDescription(LastFID, ThisFIDNumberOfMatches);

	    *RP = (t_Answer *) emalloc("GetFiles Answer", sizeof(t_Answer));
	    (*RP)->Answer = p;
	    RP = &(*RP)->Next;
	    *RP = (t_Answer *) 0;
	    ThisFIDNumberOfMatches = 0L;
	} else {
	    ++ThisFIDNumberOfMatches;
	}

	LastFID = MP->Match->Where->FID;
    }

    if (ThisFIDNumberOfMatches) {
	char *p = MakeOneDescription(LastFID, ThisFIDNumberOfMatches);

	*RP = (t_Answer *) emalloc("GetFiles Answer", sizeof(t_Answer));
	(*RP)->Answer = p;
	RP = &(*RP)->Next;
	*RP = (t_Answer *) 0;
	ThisFIDNumberOfMatches = 0L;
    }

    return Result;
}

PRIVATE char *
MakeOneDescription(FID, NumberOfMatches)
    t_FID FID;
    unsigned long NumberOfMatches;
{
    extern char *ctime();
    extern t_FileInfo *GetFileInfo();

    char *Date;
    char *p;
    t_FileInfo *FileInfo;
    char NumBuf[20];

    if (!FID) return (char *) 0;

    if (!(FileInfo = GetFileInfo(FID))) return (char *) 0;

    Date = ctime(&(FileInfo->Date));
    /** Tue Oct  3 00:57:11 BST 1989 **/
    Date[10] = '\0';

    (void) sprintf(NumBuf, "%lu", NumberOfMatches);

    p = emalloc(
	"FileName for Make One Description",
	(unsigned) (strlen(FileInfo->Name) + strlen(NumBuf) + 11)
    );
    (void) sprintf(p, "%-.5s %s %s", NumBuf, Date, FileInfo->Name);
    efree(FileInfo->Name);
    efree(FileInfo);
    return p;
}

API void
ResetPhraseMatch(Phrase)
    t_Phrase *Phrase;
{
    t_PhraseItem *Word;

    if (!Phrase || !Phrase->Words) return;

    for (Word = Phrase->Words; Word; Word = Word->Next) {
	Word->SearchIndex = 0;
    }
    Phrase->NumberOfMatches = 0;
}

/* Default is to check case, etc. only if given in the input phrase.
 * This is an enum from phrase.h, and only used in MakeMatches().
 */

extern t_PhraseCaseMatch PhraseMatchLevel;

typedef enum {
    e_First, e_Middle, e_Last
} t_Position;

API int
NumberOfWordsInPhrase(Phrase)
    t_Phrase *Phrase;
{
    register t_PhraseItem *Word;
    register int Result = 0;

    for (Word = Phrase->Words; Word; Word = Word->Next) {
	++Result;
    }
    return Result;
}

static t_FID *FilesInShortestEntry = 0;
static t_FID *FilesInShortestEntryPointer;
static long PlacesAccepted;
static long FilesInShortestEntryDeleted = 0;

PRIVATE int
AcceptWordFunc(WID, WordPlace)
    t_WID WID;
    t_WordPlace *WordPlace;
{
    t_FID LastPlausible = (*FilesInShortestEntryPointer);

    while (*FilesInShortestEntryPointer != (t_FID) 0) {
	if (*FilesInShortestEntryPointer < WordPlace->FID) { /* not there yet */
	    /* this FID is not wanted here, and therefore is not
	     * part of the phrase at all, so we'll delete it.
	     * Note that the first time through the loop we're looking
	     * either at the first FID in the array, or at the FID that
	     * matched the previous time we were called, which we must
	     * be careful to keep.  Hence LastPlausible.
	     */
	    *FilesInShortestEntryPointer = LastPlausible;
	    ++FilesInShortestEntryPointer;
	    ++FilesInShortestEntryDeleted;
	} else {
	    break;
	}
    }

    if (*FilesInShortestEntryPointer > WordPlace->FID) {
	/* gone past it */
	return 0;
    } else if (*FilesInShortestEntryPointer == WordPlace->FID) {
	++PlacesAccepted;
	return 1;
    } else {
	/* got to the end! */
	return 0;
    }
}

API long
MakeMatchesWhere(Phrase, AcceptFunction)
    t_Phrase *Phrase;
    int (*AcceptFunction)();
{
    /* Each word has a pointer (SearchIndex) to the last Word Place
     * that was examined.  This enables an O(NumberOfWords) search instead
     * of O(NumberOfWords * NumberOfWords) search.
     */
    static int ContinuesMatch();

    unsigned long PIFW; /* Place In First Word */
    t_MatchList **MLPP = &(Phrase->Matches);
    t_Match **MPP;
    t_Match **OldMatch;
    t_WordPlace *pp;
    t_PhraseItem *Word;
    t_PhraseItem *PreviousWord = 0;
    long Result = 0L;
    t_PhraseItem *LeastWord;
    int HowGood;
    long LengthOfFileArray = 0L;

    if (!Phrase) {
	return 0L;
    }

    ResetPhraseMatch(Phrase);
    /* Each iteration over this list either produces a match or rejects a
     * possible phrase starting place.
     */

    if (AsciiTrace > 1) {
	fprintf(stderr, "Match(%s)\n", Phrase->ModifiedString);
    }

    /* A phrase with garbage words can't match anything */
    if (Phrase->HasUnknownWords && PhraseMatchLevel != PCM_AnyCase) {
	return 0L;
    }

    /* Find the word in the phrase with least matches: */
    LeastWord = Phrase->Words;
    for (Word = Phrase->Words; Word; Word = Word->Next) {

	if (Word->Word->NumberOfWordPlaces <
					LeastWord->Word->NumberOfWordPlaces) {
	    LeastWord = Word;
	}
    }

    /* for efficiency, let's see which files we have to fetch */
    if (FilesInShortestEntry) {
	efree((char *) FilesInShortestEntry);
    }

    FilesInShortestEntry = (t_FID *) emalloc(
	"FIDS in Phrase",
	sizeof(t_FID) * (LeastWord->Word->NumberOfWordPlaces + 1)
	/* allow +1 for trailing 0 */
    );

    /* check we have the postings: */
    if (LeastWord->Word->WordPlacesInHere<LeastWord->Word->NumberOfWordPlaces){
	extern t_WordPlace *GetWordPlaces();
	t_WordInfo *W = LeastWord->Word; /* less indirection! */

	if (W->WordPlaces) {
	    (void) efree(W->WordPlaces);
	}

	W->WordPlaces = GetWordPlaces(
	    W->WID,
	    W->WordPlaceStart,
	    (unsigned) WIDBLOCKSIZE - (W->WordPlaceStart - W->DataBlock),
	    W->Offset,
	    W->NumberOfWordPlaces
	);
	W->WordPlacesInHere = W->NumberOfWordPlaces;
    }

    /* now complete the list of files that are possible candidates */
    {
	register long i, j;
	t_WordInfo *W = LeastWord->Word; /* less indirection! */

	FilesInShortestEntry[i = 0] = W->WordPlaces[0].FID;
	for (j = 1; j < W->NumberOfWordPlaces; j++) {
	    if (W->WordPlaces[j].FID != FilesInShortestEntry[i]) {
		i++;
		FilesInShortestEntry[i] = W->WordPlaces[j].FID;
	    }
	}
	FilesInShortestEntry[++i] = (t_FID) 0;
	LengthOfFileArray = i + 1;
	FilesInShortestEntry = (t_FID *) erealloc(
	    FilesInShortestEntry,
	    LengthOfFileArray * sizeof(t_FID)
	);
    }

    /* Ensure that the matches have been read */
    for (Word = Phrase->Words; Word; Word = Word->Next) {
	if (Word->Word->WordPlacesInHere < Word->Word->NumberOfWordPlaces) {
	    extern t_WordPlace *GetWordPlacesWhere();
	    t_WordInfo *W = Word->Word; /* less indirection! */

	    if (W->WordPlaces) {
		(void) efree(W->WordPlaces);
	    }

	    FilesInShortestEntryPointer = FilesInShortestEntry;
	    PlacesAccepted = 0L;
	    W->WordPlaces = GetWordPlacesWhere(
		W->WID,
		W->WordPlaceStart,
		(unsigned) WIDBLOCKSIZE - (W->WordPlaceStart - W->DataBlock),
		W->Offset,
		W->NumberOfWordPlaces,
		AcceptWordFunc
	    );
#ifdef ASCIITRACE
	    if (AsciiTrace > 4) {
		fprintf(stderr,
		    "%s: %ld/%ld wanted\n",
		    W->Word,
		    PlacesAccepted,
		    W->NumberOfWordPlaces
		);
	    }
#endif
	    W->WordPlacesInHere = W->NumberOfWordPlaces = PlacesAccepted;

	    if (!PlacesAccepted) {
		return (long) 0;
	    }

	    /* spot trailing unwanted documents: */
	    if (*FilesInShortestEntryPointer) {
		if (*++FilesInShortestEntryPointer) {
		    *FilesInShortestEntryPointer = (t_FID) 0;
		    FilesInShortestEntryDeleted += (
			LengthOfFileArray +1 -
			(FilesInShortestEntryPointer - FilesInShortestEntry)
		    );
		    /* (+1 for the trailing zero) */
		}
	    }

	    if (Word->Next && FilesInShortestEntryDeleted > 1) {
		register t_FID *WhereWeAreAt;
		register t_FID *This;

		WhereWeAreAt = This = FilesInShortestEntry;

		while (*This) {
		    if (*This == This[1]) {
			t_FID Bad = (*This);

			while (*This == Bad) {
			    This++;
			}
			continue;
		    }

		    if (This != WhereWeAreAt) {
			*(WhereWeAreAt++) = *(This++);
		    } else {
			++WhereWeAreAt;
			++This;
		    }
		}

		*WhereWeAreAt = (t_FID) 0;
		LengthOfFileArray = (WhereWeAreAt - FilesInShortestEntry) + 1;
		FilesInShortestEntry = (t_FID *) erealloc(
		    FilesInShortestEntry,
		    LengthOfFileArray * sizeof(t_FID)
		);
	    }
	}
    }

    /* For each match in the first word in the phrase: */
    for (PIFW = 0; PIFW < Phrase->Words->Word->NumberOfWordPlaces; PIFW++) {
	t_WordPlace *LastFOP = (t_WordPlace *) 0;

	/* The idea is that the next two loops are are likely to reduce
	 * considerably the number of places we have to consider in the
	 * case that the first word in the phrase has a lot of matches
	 * and there is a subsequent word with relatively few matches.
	 * Experiments suggest that this is fairly common.
	 *
	 * This is still a nearly (i.e. slightly-better-than) linear
	 * algorithm w.r.t the total number of matches in all of the
	 * words added up.  Note that I alter LeastWord->SearchIndex in
	 * one of the two loops that follow, so when WordPlaces from that
	 * word are considered, we don't have to look at any twice.
	 *
	 * In order to do better, one would have to be able to avoid
	 * looking at some or (better!) most of the WordPlaces.
	 *
	 * For example, not fetching so many from disk:
	 * if we didn't do the fetches until we needed to, and we gave
	 * GetWordPlaces a minimum FID to look for, we might be able
	 * to reduce things by (say) 15%.
	 * If all of the FIDS were stored separately, we would not
	 * have to look at the (Block, Word, Flags, StuffBefore) stuff at
	 * all, and that would be much faster.  One way to do that might be
	 * to store the list of FIDs with the word (as now), and perhaps
	 * some flags and the count of words/fid, and to store the rest
	 * in a per-file data structure.
	 *
	 * That would be a major, major hack...
	 * 					... sigh.
	 */

	while (LeastWord->Word->WordPlaces[LeastWord->SearchIndex].FID <
		    Phrase->Words->Word->WordPlaces[PIFW].FID) {
	    if (++(LeastWord->SearchIndex) >=
					LeastWord->Word->NumberOfWordPlaces) {
		goto GiveUp;
	    }
	}

	while (Phrase->Words->Word->WordPlaces[PIFW].FID <
		LeastWord->Word->WordPlaces[LeastWord->SearchIndex].FID) {
	    if (++PIFW >= Phrase->Words->Word->NumberOfWordPlaces) {
		goto GiveUp;
	    }
	}
	/* The following comment tells Sabre_C not to moan about "if (0)" */
	/*SUPPRESS558*/
	if (0) {
GiveUp:
	    break;
	}
	/* end of attempted speed improvement */


	MPP = (t_Match **) 0;
	pp = &Phrase->Words->Word->WordPlaces[Phrase->Words->SearchIndex = PIFW];
	/* When we have a partially completed match,
	 * FOP (declared below) will point to the WordPlace currently
	 * being considered to see if it extends the partial match;
	 * LastFOP points to the previous WordPlace in the match.
	 */

	/* For each word in the phrase: */
	for (Word = Phrase->Words; Word; Word = Word->Next) {
	    int GotOne = 0;

	    /* For each occurrence of that word: */
	    for (; Word->SearchIndex < Word->Word->NumberOfWordPlaces;
							++Word->SearchIndex) {
		register t_WordPlace *FOP =
			&Word->Word->WordPlaces[Word->SearchIndex];
		
		if (!LastFOP) {
		    LastFOP = FOP;
		}

		/** So:
		 ** | int PIFW = each match in the first word in the phrase
		 ** | t_WordPlace *pp = each match in the phrase
		 **    | t_PhraseItem *Word = each word in the phrase
		 **	   | unsigned SearchIndex = each match of that word
		 **	   | t_WordPlace *FOP = each occurrence of that word
		 **
		 ** Hence, we are comparing pp and FOP, hoping that each time
		 ** round the (Word) loop we will advance FOP.
		 ** Once we have decided that FOP and pp relate to the
		 ** same file and that FOP is no earlier than pp in the
		 ** file, we must then check that FOP is advancing the
		 ** chain by comparing it to the previous element in the
		 ** list (LastFOP).
		 **
		 ** When we break from this inner list, we must either have
		 ** eliminated this particular (PIFW) as starting a match-
		 ** chain, or have decided that we have extended the
		 ** current match chain (by setting GotOne).
		 **/

		if (LastFOP == FOP) { /* same word! */
		    HowGood = CheckFlags(
			Word->Word,
			FOP,
			    (Word == Phrase->Words) ? e_First : (
				( (Word->Next) ? e_Last : e_Middle)
			    )
			);
		} else if (pp->FID == FOP->FID) {
		    HowGood = ContinuesMatch(
			PreviousWord,
			Word,
			pp,
			LastFOP,
			FOP,
			(Word == Phrase->Words) ? e_First : (
			    ( (Word->Next) ? e_Last : e_Middle)
			)
		    );
		} else {
		    HowGood = (pp->FID > FOP->FID) ? -1 : 1;
		}

		switch (HowGood) {
		case 0:
		    /* G O T C H A !!!! */
		    /* extend the HitList, since it's OK so far. */

		    /* allocate a new match: */
		    if (!MPP) {
			*MLPP = (t_MatchList *)
			    emalloc("MatchList entry for MakeMatches",
			    sizeof(t_MatchList)
			);
			(*MLPP)->Match = (t_Match *) 0;
			OldMatch = MPP = &((*MLPP)->Match);
			MLPP = &(*MLPP)->Next;
			*MLPP = (t_MatchList *) 0;
		    }

		    *MPP = (t_Match *) emalloc("Phrase Match", sizeof(t_Match));
		    (*MPP)->WID = Word->Word->WID;
		    (*MPP)->Where = FOP;
		    (*MPP)->Next = (t_Match *) 0;
		    MPP = &(*MPP)->Next;
		    GotOne++;
		    break;
		case 1: /* gone too far */
#ifdef ASCIITRACE
		    if (AsciiTrace > 10) {
			t_WordInfo WW;
			
			WW = *(Word->Word);

			if (LastFOP == FOP) {
			    /* UnFlag() returns a pointer to a static buffer,
			     * so I have to use two printf() calls here.
			     */
			    fprintf(stderr, "Reject(%s (%d) != ",
				    UnFlag(&WW, WW.WordPlace.Flags),
				    WW.WordPlace.Flags);
			    fprintf(stderr, "%s (%d)) [flags]\n",
				    UnFlag(&WW, FOP->Flags), FOP->Flags);
			} else {
			    fprintf(stderr, "Reject(%s) -- too far\n",
				UnFlag(&WW, WW.WordPlace.Flags));
			}
		    }
#endif
		    break;
		case -1:
		    continue; /* not there yet */
		default:
		    Error(E_FATAL|E_INTERNAL,
			"%s: %d HowGood == %d",
			__FILE__, __LINE__,
			HowGood
		    );
		}

		/* Remember where we got up to... so that we can extend
		 * the list when we start looking at the next word.
		 */
		LastFOP = FOP;

#ifdef ASCIITRACE
		if (AsciiTrace >= 4) {
		    t_WordInfo WW;
		    
		    WW = *(Word->Word);
		    /* UnFlag() returns a pointer to a static buffer */
		    fprintf(stderr, "Partial match %s",
				UnFlag(&WW, Word->Word->WordPlace.Flags));
		    fprintf(stderr, "(Word (%s,%lu,%u) in file %lu)\n",
				    UnFlag(&WW, FOP->Flags),
				    FOP->BlockInFile, FOP->WordInBlock,
				    FOP->FID
		    );
		}
#endif
		/* If we got to here, we extended the list, which is fine;
		 * otherwise, if we hit a continue, we try to carry on
		 * looking at matches of this word, and if we hit a break
		 * before we set "GotOne", we give up on this match
		 * altogether.
		 */
		break;
	    } /* For each occurrence of that word: */

	    if (!GotOne) {
		t_Match *MP;
		/* This word isn't here, so neither is the phrase found
		 * in this file starting here.
		 */

		if (MPP) {
		    for (MP = (*OldMatch); MP != (t_Match *) 0; /*void*/) {
			t_Match *Next = MP->Next;

			efree((char *) MP);
			MP = Next;
		    }
		    *OldMatch = (t_Match *) 0;
		}

		break;
	    } else {
		/* If we've reached the end of the phrase, i.e. if
		 * Word->Next is zero, we have successfully added a new
		 * phrase!
		 */
		if (Word->Next == (t_PhraseItem *) 0) {
		    if (AcceptFunction) {
			switch ((*AcceptFunction)(Phrase, *OldMatch)) {
			case 1: /* OK */
			    Result++;
#ifdef ASCIITRACE
			    if (AsciiTrace > 4) {
				fprintf(stderr, "OK, Result now %d\n", Result);
			    }
#endif
			    break;
			case -1:
			    /* accept and give up */
#ifdef ASCIITRACE
			    if (AsciiTrace > 4) {
				fprintf(stderr,
				    "AcceptFunc returned -1, giving up at %d\n",
				    Result
				);
			    }
#endif
			    return Phrase->NumberOfMatches = Result;
			case 0:
			    /* free the match structure */
#ifdef ASCIITRACE
			    if (AsciiTrace > 4) {
				fprintf(stderr,
				    "AcceptFunc rejected match, still %d\n",
				    Result
				);
			    }
#endif
			    if (MPP) {
				t_Match *MP = (*OldMatch);

				while (MP != (t_Match *) 0) {
				    t_Match *Next = MP->Next;

				    efree((char *) MP);
				    MP = Next;
				}
			    }

			    *OldMatch = (t_Match *) 0;
			}
		    } else {
			++Result;
#ifdef ASCIITRACE
			if (AsciiTrace > 4) {
			    fprintf(stderr, "Result now %d\n", Result + 1);
			}
#endif
		    }
		}
	    }

	    PreviousWord = Word;

	} /* end for (each word in the phrase) */
    } /* end (for each FID/Offset pair in the first word */
    return Phrase->NumberOfMatches = Result;
}

API long
MakeMatches(Phrase)
{
    return MakeMatchesWhere(Phrase, (int (*)())0);
}

PRIVATE int
ContinuesMatch(PreviousWord, Word, First, Prev, New, Position)
    t_PhraseItem *PreviousWord;
    t_PhraseItem *Word;
    register t_WordPlace *First;
    t_WordPlace *Prev;
    register t_WordPlace *New;
    t_Position Position;
{
    register t_WordInfo *QueryWord = Word->Word;

    /* Return Value is
     *  -1 --- if New occurs before Prev (and thus isn't part of the match)
     *  0  --- if it's the next word in the match
     *  +1 --- if we've gone past it
     * Note: you can use these values in a switch() if you want.
     */

    /* First check we are looking at the right file:
     * Have we gone far enough?
     */
    if (New->FID < First->FID) {
	return -1; /* not far enough */
    } else if (New->FID > First->FID) {
	return 1; /* too far */
    } else if (Prev == New) {
	return 0;
    }

    /* Hey everybody, they're the same!
     * That means that this might be a candidate for a MATCH!!!!
     */
    
    /* if (SimplyAnywhereWillDo) { OK; break; } */

    /* Clearly later words in the phrase can't be in earlier
     * blocks...
     */
    if (New->BlockInFile < First->BlockInFile) {
	/* Although we are in the right file, we have not
	 * yet reached the correct offset.
	 */
	return -1;
    } 

    /* If we get to here,
     * . we are in the right file
     * . we are at least as far into the file as the start
     *   of the phrase
     */

    /* Now check that we are a reasonable distance past
     * the preceding word (checking that we are not on the first
     * match in the list, of course):
     */
    if (New->BlockInFile < Prev->BlockInFile) {
	/* not gone far enough */
	return -1;
    }
    if (New->BlockInFile > Prev->BlockInFile + 1) {
	/* If they are more than one block apart, I
	 * don't believe them to be part of a phrase!
	 */
	return 1;
    }

    if (New->BlockInFile == Prev->BlockInFile) {
	int DistanceApartInQuery = 1;

	if (PreviousWord && PreviousWord->Word) {
	    DistanceApartInQuery = Word->Word->WordPlace.WordInBlock -
				PreviousWord->Word->WordPlace.WordInBlock;
	}

	/* If they are in the same block, one must be
	 * exactly one word beyond the other.  I don't
	 * think they can ever be the same, unless there
	 * is a bug somewhere!  (Query expansion and anaphoristic
	 * reference handling might change that in the future, though!)
	 *
	 * However, if common words were skipped in the query,
	 * there may be a gap.
	 *
	 * In principle, it's possible for a "common word" -- e.g. a number --
	 * to span a whole block, and then we can't match words on both
	 * sides of it.  Variable sized blocks will cure this, as then a
	 * phrase will never span a block.
	 */
	if (New->WordInBlock <= Prev->WordInBlock) {
	    /* too early in the block */
	    return -1;
	}
	switch (PhraseMatchLevel) {
	case PCM_AnyCase:
	    if (New->WordInBlock > 
				Prev->WordInBlock + DistanceApartInQuery + 2) {
		return 1; /* gone too far */
		/* We allow a few words slop in this case, though... */
	    }
	    break; /* clearly OK */
	case PCM_SameCase:
	case PCM_HalfCase:
	default:
	    if (New->WordInBlock > Prev->WordInBlock + DistanceApartInQuery) {
		return 1; /* gone too far */
	    }
	}
    } else {
	/* they are in adjacent blocks */
	if (New->WordInBlock > 0 ||
			!(Prev->Flags & WPF_LASTINBLOCK)) {
	    /* there is another word between them, so
	     * we have gone too far.
	     * I went to a lot of effort in addfile.c to
	     * mantain that flag, just for this!
	     */
	    return 1;
	}
    }
    
    /* Now we check that the word matches the given word
     * -- in other words, that possessive/plural/case
     *    is correct if required.  Do this later as it is
     *    relatively expensive I expect, and we will not
     *    usually care about case.
     *
     * Since the word is in the right place, if it fails here there
     * is no point in looking at the next word in this block!
     */

    return CheckFlags(QueryWord, New, Position);
}

PRIVATE int
CheckFlags(QueryWord, New, Position)
    t_WordInfo *QueryWord;
    t_WordPlace *New;
    t_Position Position;
{
#ifdef ASCIITRACE
    if (AsciiTrace > 4) {
	fprintf(stderr, "CheckFlags %s (%s Qf=%d cmp f=%d) ",
	    QueryWord->Word,
	    Position == e_First ? 
		"First" : ( (Position == e_Middle ) ? "Middle" : "Last" ),
	    QueryWord->WordPlace.Flags, New->Flags);
    }
#endif

    /* Phrase matching levels currently have silly names, sorry. */
    switch (PhraseMatchLevel) {

    default: /* defensive! */
	Error(E_FATAL|E_INTERNAL|E_BUG,
	    "%d %s: unknown phrase match level %d",
	    __LINE__, __FILE__,
	    PhraseMatchLevel
	);
    case PCM_AnyCase:
	break; /* clearly OK */
    case PCM_SameCase:
	if (QueryWord->WordPlace.StuffBefore > 0 || Position != e_First) {
	    int Difference;

	    Difference = QueryWord->WordPlace.StuffBefore - New->StuffBefore;
	    if (Difference < -2 || Difference > 4) {
		return 1; /* give up on this match */
	    }
	}

	/** for things before the current word, be "heuristic" if
	 ** it's the first word in the phrase, as the user can't have
	 ** known about the context of the putative match, only its contents.
	 **/

	/* Skipped common words: */
	if (Position == e_First) {
	    /* It's the first word in the query, so only check for things
	     * before this word if they were specified:
	     */
	    if ((QueryWord->WordPlace.Flags & WPF_LASTWASCOMMON) &&
			    !(New->Flags & WPF_LASTWASCOMMON)) {
		return 1;
	    }

	    if ((QueryWord->WordPlace.Flags & WPF_LASTHADLETTERS) &&
				!(New->Flags & WPF_LASTHADLETTERS)) {
		return 1; /* give up on this match */
	    }

	    if ((QueryWord->WordPlace.Flags & WPF_LASTHADPUNCT) &&
				!(New->Flags & WPF_LASTHADPUNCT)) {
		return 1; /* give up on this match */
	    }

	} else { /* not the first word */
	    if ((New->Flags &
		    (WPF_LASTWASCOMMON|WPF_LASTHADLETTERS|WPF_LASTHADPUNCT)) !=
		(QueryWord->WordPlace.Flags & 
		    (WPF_LASTWASCOMMON|WPF_LASTHADLETTERS|WPF_LASTHADPUNCT))) {
		return 1; /* give up on this match */
	    }
	}

	/* All the non-context-sensitive flags at once: */
	if ((QueryWord->WordPlace.Flags & (WPF_WASPLURAL|WPF_UPPERCASE)) != 
			(New->Flags & (WPF_UPPERCASE|WPF_WASPLURAL))) {
	    /* The cases are different, no good */
	    return 1; /* give up on this match */
	}

	/* Finally, the least reliable, so always heuristic: */
	if ((QueryWord->WordPlace.Flags & WPF_POSSESSIVE) &&
				!(New->Flags & WPF_POSSESSIVE)) {
	    return 1; /* give up on this match */
	}

	break;
    case PCM_HalfCase:
	/* In this case, we are lax about things, but if the
	 * user typed  plural/possessive/capital, we only
	 * match one with the same attribute.
	 */
	if ((QueryWord->WordPlace.Flags & WPF_UPPERCASE) &&
				    !(New->Flags & WPF_UPPERCASE)) {
#ifdef ASCIITRACE
	    if (AsciiTrace > 4) {
		fprintf(stderr, "Reject [uppercase]\n");
	    }
#endif
	    return 1; /* give up on this match */
	}

	/* plurals: this should be separate */
	if ((QueryWord->WordPlace.Flags & WPF_WASPLURAL) &&
	    !(New->Flags & WPF_WASPLURAL)) {
#ifdef ASCIITRACE
	    if (AsciiTrace > 4) {
		fprintf(stderr, "Reject [plural]\n");
	    }
#endif
	    return 1; /* give up on this match */
	}

	/* Now, what about skipped common words? */
	if ((QueryWord->WordPlace.Flags & WPF_LASTWASCOMMON) &&
				!(New->Flags & WPF_LASTWASCOMMON)) {
#ifdef ASCIITRACE
	    if (AsciiTrace > 4) {
		fprintf(stderr, "Reject [last was common]\n");
	    }
#endif
	    return 1; /* give up on this match */
	}

	/* Stuff before, if given, must be present: */
	if (QueryWord->WordPlace.StuffBefore > 1) {
	    if (New->StuffBefore < QueryWord->WordPlace.StuffBefore - 1) {
#ifdef ASCIITRACE
		if (AsciiTrace > 4) {
		    fprintf(stderr, "Reject [Stuff Before %d != Q%d]\n",
			    QueryWord->WordPlace.StuffBefore,
			    New->StuffBefore);
		}
#endif
		return 1;
	    } /* don't care if there is too much there, though */
	}

	if ((QueryWord->WordPlace.Flags & WPF_LASTHADPUNCT) &&
				   !(New->Flags & WPF_LASTHADPUNCT)) {
#ifdef ASCIITRACE
	    if (AsciiTrace > 4) {
		fprintf(stderr, "Reject [punctuation]\n");
	    }
#endif
	    return 1; /* give up on this match */
	}

	if ((QueryWord->WordPlace.Flags & WPF_POSSESSIVE) &&
				   !(New->Flags & WPF_POSSESSIVE)) {
#ifdef ASCIITRACE
	    if (AsciiTrace > 4) {
		fprintf(stderr, "Reject [user flag]\n");
	    }
#endif
	    return 1; /* give up on this match */
	}
	break;
    }

    /* If we got here...
     *
     */
    
    return 0; /* It's all OK! */
}

