/* matchend.c -- Copyright 1993 Liam R. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

/* matchend -- find the start & end bytes of a match
 * Liam Quin, September 1993 and later...
 *
 * $Id: matchend.c,v 1.5 93/11/28 15:16:42 lee Exp $ 
 */

#include "error.h"
#include <stdio.h>
#include <malloc.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef BSD
# include <strings.h>
#else
# include <string.h>
#endif

#include "globals.h" /* defines and declarations for database filenames */

#include "fileinfo.h"
#include "wordinfo.h"
#include "wordrules.h"
#include "emalloc.h"
#include "addfile.h"
#include "wmoffset.h"
#include "Revision.h"

#ifdef SYSV
#define TOLOWER(ch) ch = tolower(ch)
#else
#define TOLOWER(ch) if (isupper(ch)) ch = tolower(ch)
#endif

extern t_WordInfo *SpReadWord(/* char **Stringpp, **Startp */);

/** System calls and library routines used in this file: **/
/** System calls: **/
/** Library Functions: **/
#ifndef tolower
 extern int tolower();
#endif

/** Functions within this file that need declaring: **/

#ifdef ASCIITRACE
extern int AsciiTrace;
#endif

/**/

t_OffsetPair *
FindMatchEnds(Buffer, Length, StartBlock, BIF, WIB, NumberOfWords)
    char *Buffer;
    unsigned int Length;
    char *StartBlock;
    int BIF;
    int WIB;
    int NumberOfWords;
{
    static t_OffsetPair Result;
    int WordsSeen = 0;

    /* Find the start and end of a match */
    char *p; /* not register, we take its address */

    (void) SpReadWord((char **) 0, (char *) 0, 0); /* initialise */
    p = StartBlock; 

    /* This is complicated because if a word crosses a block
     * boundary, we would see just the end of it and think it
     * the first word in the block.
     * Hence, if it goes backwards, we'll skip over it.
     */

#define PossibleWordCharacter(c) \
    (EndsWord(c)||StartsWord(c)||isdigit(c)||(c == '.')||OnlyWithinWord(c))

    if (
	p > Buffer &&
	!isspace(*p) &&
	!isspace(p[-1]) && 
	PossibleWordCharacter(*p) && PossibleWordCharacter(p[-1])
    ) {
	/* There was a word spanning the block boundary.
	 * Look at the previous block, and see if we can handle that more
	 * easily:
	 */
	if ((BIF == 1) || (
			    (p - Buffer > FileBlockSize) && (
				isspace(p[-FileBlockSize]) ||
				isspace(p[-(FileBlockSize - 1)]) ||
				!PossibleWordCharacter(p[-FileBlockSize]) ||
				!PossibleWordCharacter(p[-(FileBlockSize - 1)])
			    )
			  )
	) {
	    char *q;
	    t_WordInfo *WP;
	    char *Start;

	    /* Yes, there's a word boundary at the start of the
	     * previous block
	     */

	    Start = q = &p[-FileBlockSize];

	    /* find the start of the block */
	    for (;;) {
		WP = SpReadWord(&q, &Start, 0); /* 0=don't ignore common words */
		if (WP && WP->WordPlace.BlockInFile > 0) {
		    break;
		}
	    } while (WP && Start < p);

	    if (WP) {
		p = Start; /* or q? */
	    }
#ifdef ASCIITRACE
	    if (AsciiTrace > 2) {
		(void) fflush(stderr);
		(void) fflush(stdout);
		printf("<<1>>");
	    }
#endif
	} else {
	    char *q;
	    t_WordInfo *WP;
	    char *Start = p;

	    for (q = p; q > Buffer; q--) {
		if (isspace(*q) /* || !PossibleWordCharacter(*q)*/) break;
	    }

	    do {
		WP = SpReadWord(&q, &Start, 0);
	    } while (WP && Start < p);

	    if (WP) {
		p = Start;
	    }
#ifdef ASCIITRACE
	    if (AsciiTrace > 2) {
		(void) fflush(stderr);
		(void) fflush(stdout);
		printf("<<1>>");
	    }
#endif
	}

	(void) SpReadWord((char *) 0, (char *) 0, 0); /* initialise again */

    }

    Result.Start = Result.End = (char *) 0;

    for (;;) {
	t_WordInfo *WP;

	if (!p || !*p) {
	    return (t_OffsetPair *) 0;
	}

	WP = SpReadWord(&p, &Result.Start, 1);

	if (!WP) {
	    return (t_OffsetPair *) 0;
	}

	if (WP->WordPlace.WordInBlock >= WIB) {
	    break;
	}
    }

    /* We have found the first word in the match;
     * now find the end of the last word:
     */

    WordsSeen = 1;
    while (p && *p && WordsSeen < NumberOfWords) {
	t_WordInfo *WP;

	WP = SpReadWord(&p, (char *) 0, 0);
	if (!WP) break;

	++WordsSeen;
    }
    Result.End = --p;
    return &Result;
}
