/* IsNZFile.c -- Copyright 1993 Liam R. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

/* $Id: isnzfile.c,v 1.1 93/12/08 23:06:47 lee Exp $ */

#include <sys/types.h>
#include <sys/stat.h>

#ifdef ASCIITRACE
# include <stdio.h>
extern int AsciiTrace;
#endif

/* IsNZFile(Path) - return 1 if the given Pathname represents a regular file
* of non-zero length, and 0 otherwise.  
 *
 * It's tempting (and wrong) to write IsFile as
 *	return !IsDir(Path);
 * but this fails for special devices (tape drives, &c.).
 *
 * There is tracing in here so that you can see which files are being
 * investigated by the calling program... On systems that have the `trace'
 * utility, or strace or truss(1), investigate using that instead.
 */

int
IsNZFile(Path)
    char *Path;
{
    struct stat statbuf;
    register int Result = 1;

#ifdef ASCIITRACE
    if (AsciiTrace > 20) {
	fprintf(stderr, "IsNZFile(%s)...", Path);
    }
#endif
    if (stat(Path, &statbuf) < 0) return 0;
    if ((statbuf.st_mode & S_IFMT) != S_IFREG ||
	(statbuf.st_size == 0)) {
	Result = 0;
    }
#ifdef ASCIITRACE
    if (AsciiTrace > 20) {
	fprintf(stderr, " %s\n", Result ? "true" : "false");
    }
#endif
    return Result;
}
