#! /bin/sh
: use /bin/sh
# put the : line first on System V

# $Id: lq.sh,v 1.15 93/12/08 23:03:24 lee Exp $

# PATH should include mawk and tput.
# If you don't have mawk, it's a much faster version of awk --
# pick it up!  If you don't have tput, there are public domain
# versions around, or you can do
#
# tput()
# {
# }
#
# to disable the use of bold, underscore, etc.
#
#
# Finally, "rs" is included in the lq-text distribution in the
# src/tools directory.
#

PATH=/usr/local/bin:/usr/5bin:$PATH:/home/sqarc/lee/bin/sun4os4
export PATH

RS=rs
export RS

N=; C='\c'; if test x"`echo -n hello`" = x"hello" ; then N='-n'; C=; fi
export N C
# Single Quote
Q="'"

COLS=`tput cols 2>/dev/null`
if [ x"$COLS" = x"" ]
then
    COLS=80
fi

LEFT=`expr '(' $COLS - 5 ')' / 2`
RIGHT=$LEFT

if test -z "$CMDNAME"
then
    CMDNAME="`basename $0`"
else
    CMDNAME="${CMDNAME}: `basename $0`"
fi
export CMDNAME
PS1="[${CMDNAME}] $PS1" 
PS2="[${CMDNAME}] $PS2" 
export PS1 PS2

lqphrase=lqrank
export lqphrase

# commands:
SETCMD=":set"
VIEWCMD=":view"
INDEX=":index"
FILES=":files"
HELPCMD=":help"
LESSCMD=":less"
PAGECMD=":page"
WORDLIST=":prefix"
WORDGREP=":grep"
SHELL_CMD=":shell"
PAGER="${PAGER-more}"
# export SETCMD VIEWCMD INDEX FILES HELPCMD LESSCMD PAGECMD WORDGREP WORDLIST PAGER SHELL_CMD


if test -x /usr/local/bin/less -o -x /usr/ucb/less
then
    LESSPROG=less
else
    LESSPROG=${PAGER-more}
fi

LQTEXTDIR="${LQTEXTDIR-$HOME/LQTEXTDIR}"
export LQTEXTDIR
echo "Using database in $LQTEXTDIR..."

# if we have tput, use it; if not, forget it:
if (tput rmso > /dev/null) 2> /dev/null
then
    # Yes, tput works OK
    smso="`tput smso`"; rmso="`tput rmso`"
    smul="`tput smul`"; rmul="`tput rmul`"
    bold="`tput bold`"; rmall="`tput sgr0`"
    rmbold="${rmall}"
    if [ x"${smul}" = x"" ]
    then
	smul="${smso}"; rmul="${rmso}"
    elif [ x"${smso}" = x"" ]
    then
	smso="${smul}"; rmso="${rmul}"
    fi
    if [ x"$bold}" = x"" ]
    then
	VBAR="${smso}|${rmso}"
	bold="${smul}"
	rmbold="${rmul}"
	if [ x"${rmall}" = x"" ]
	then
	    rmall="${rmso}"
	fi
    else
	VBAR="${bold}|${rmbold}"
    fi
else
    smso=""; rmso=""
    smul="_"; rmul="_"
    bold=""; rmbold=""
fi
lqtext="${smul}lq-text${rmul}"

set_traps()
{
    trap "/bin/rm -f $tmp;echo ${rmso}${rmul}${rmall}Quit.; exit" 0 1 15
    trap "/bin/rm -f $tmp;echo ${rmso}${rmul}${rmall}interrupted, bye;exit" 2 3
}

unset_traps()
{
    trap "/bin/rm -f $tmp;echo ${rmso}${rmul}${rmall}Quit.; exit" 0 1 15
    trap : 2 3
}

unset_all_traps()
{
    trap : 0 1 2 3 15
}

set_traps

quit=no

SGML_CHAR=""

if [ x"$1" = x"-sgml" ]
then
    SGML_CLEVER="-c -C ${SGML_CHAR}"
    SGML_TAG_DESC="as ${SGML_CHAR} characters"
    ENTLIST="-E ${LQTEXTDIR}/entitylist"
else
    SGML_CLEVER=
    SGML_TAG_DESC="as ASCII"
    ENTLIST=
fi
export SGML_TAG_DESC SGML_CLEVER ENTLIST

prog=lqkwic
export prog
PRECISION=;
PRECISION_DESC="heuristic matching (the default)"
RANK="all";
RANK_TYPE_DESC="find documents containing ${smul}all${rmul} of the phrases"
PROMPT="${bold}>${rmbold} "
export PROMPT
MAXHITS="200"
export MAXHITS

print_info()
{
    if test x"$1" != x""
    then
	echo $N "$@$C"
    else
	cat
    fi |
    sed -e "s/^[ 	]*|/${VBAR}/" \
	-e "/^[ 	]*#/d" \
	-e "s/@</${smul}/g" \
	-e "s/@>/${rmul}/g" \
	-e "s/@\[/${bold}/g" \
	-e "s/@\]/${rmbold}/g"
}

main_help()
{
    print_info "\
    | Type phrases one per line, for example
    | @<happiness and good fortune@>
    | and, when you are done, press @<return@> twice.  ${CMDNAME} will then
    | ${RANK_TYPE_DESC}.
    | There are various options you can set;
    | type @<${SETCMD}@> to see them.
    | Once you've seen the index, you can use the following commands:
    | ${VIEWCMD}, ${INDEX}, ${FILES}, ${HELPCMD}, ${LESSCMD}, ${PAGECMD}, ${WORDLIST}, ${WORDGREP}
    | 
"
}

short_index_help()
{
    print_info "\
    | Use @<${INDEX}@> to see the index again, @<${HELPCMD}@> for other options.
"
}

command_help()
{
    print_info "\
    | @[Commands@]
    |
    | @[${HELPCMD}@]
    |    Shows you this message.
    | @[${VIEWCMD}@] [@<n@>,@<n1-n2@>,@<n@>...]
    |    Use @<${VIEWCMD}@> to see the text surrounding matches.
    |    The number @<n@> is from the left-hand edge of the index;
    |    @[${SETCMD} maxhits@] explains @<range@>s in more detail.
    | @[${PAGECMD}@] [@<n@>,@<n1-n2@>,@<n@>...]
    |    Uses ${PAGER} (which you can set
    |    in the \$PAGER Unix environment variable) to show the files matched.
    | @[${LESSCMD}@] [@<n@>,@<n1-n2@>,@<n@>...]
    |    The same as @[${PAGECMD}@] except that it starts on the line number
    |    containing the match, firing up the pager separately on
    |    each file.
    | @[${INDEX}@] [@<n@>,@<n1-n2@>,@<n@>...]
    |    This shows the indexfor the phrases you've typed.
    | @[${FILES}@] [@<n@>,@<n1-n2@>,@<n@>...]
    |    This simply lists all of the files that were matched,
    |    in ranked order.
    | @[${WORDLIST}@] @<prefix-string@>
    |    Shows all of the words in the database that begin with that prefix,
    |    together with the number of times they occur.
    | @[${WORDGREP}@] @<egrep-pattern>
    |    Shows all of the words in the database that match that egrep pattern,
    |    together with the number of times they occur (not the number of files
    |    in which they appear, though).
    | @[${SETCMD}@] @<option@> ...
    |    Type @<${SETCMD}@> to see more information about setting options.
    | @[${SHELL_CMD}@] [@<command@> ...]
    |    Use ${SHELL-sh} to run commands.
    |
    | Commands that take a list of matches (@<n@>,@<n1-n2@>,@<n@>...)
    | only work when you have generated a list of matches.  If you don't give
    | any arguments, you get the whole list.
    |
"
}

make_match_list()
{
    # $1 is output file
    tmp="$1"
    shift

    if [ x"$1" = x"" ]
    then
	echo "p" >> $tmp
    elif expr "$1" : all > /dev/null
    then
	echo "p" >> $tmp
    else
	echo "$@" | tr ', ' '\012\012' | mawk -F- '
	    (NF == 2) {
		if ($2 < $1) {
		    tmp = $1
		    $1 = $2
		    $2 = tmp
		}
		if ($1 < 1) $1 = 1
		printf "%d,%dp\n", $1, $2
		next
	    }
	    {
		printf "%dp\n", $1
	    }
	' >> $tmp
    fi
}

handle_shell()
{
    shift
    unset_all_traps
    if test -z "$1"
    then
	${SHELL-/bin/sh}
    else
	${SHELL-/bin/sh} -c "$@"
    fi
    print_info "\
    |
    | (${CMDNAME} -- continue)
"
    set_traps
}

# cf. handle_page below...
handle_less()
{
    # handle_less ${LESSCMD} n m [...]

    shift

    unset_traps

    case "$*" in
    *[a-zA-Z0-9,-]*)
	;;
    *)
	set all
	;;
    esac

    echo Use control-C q to stop paging...
    eval $lqphrase ${PRECISION} -r ${RANK} $Phrases |
	lqkwic -S '' -s '${StartByte} ${FileName}${NL}' -o "$*" |
	lqbyteline -f - | {
	    while read line file
	    do
		${LESSPROG} +$line "$file" < /dev/tty
	    done
	}


    set_traps

    short_index_help
}

handle_page()
{
    # view ${VIEWCMD} n m [...]
    shift

    # now append the sed script:

    case "$*" in
    *[a-zA-Z0-9,-]*)
	;;
    *)
	set all
	;;
    esac


    unset_traps

    eval $lqphrase ${PRECISION} -r ${RANK} $Phrases |
	lqkwic -o "$*" -S '' -s '${FileName}${NL}' |
	xargs ${PAGER}

    set_traps

    short_index_help
}

handle_view()
{
    # view ${VIEWCMD} n m [...]
    shift

    tmp=/tmp/lq.$$
    rm -f $tmp
    make_match_list $tmp "$@"

    unset_traps

    eval $lqphrase ${PRECISION} -r ${RANK} $Phrases |
	sed -n -f $tmp | lqshow -f -
    set_traps
    rm -f $tmp
    echo
    short_index_help
}

show_index()
{
    shift

    unset_traps
    eval $lqphrase ${PRECISION} -r ${RANK} $Phrases |
    lqkwic -e ${ENTLIST} -p 1 -g 0 -l$LEFT -r$RIGHT $SGML_CLEVER -A \
    -s \
'${MatchNumber/3r}:${TextBefore}'"${bold}"'${MatchedText}'"${rmbold}"'${TextAfter}${NL}' \
    -S "${smul}"'${DocTitle}'"${rmul}"' == ${DocName}${NL}' \
    -f - # "${@-\"1-${MAXHITS}\"}"
    set_traps
    rm -f $tmp
}

show_text()
{
    tmp=/tmp/lq.$$
    rm -f $tmp

    make_match_list $tmp "$@"

    unset_traps
    eval $lqphrase ${PRECISION} -r ${RANK} $Phrases | sed -n -f $tmp |
    $prog -f -
    set_traps
    short_index_help
}

show_filenames()
{
    unset_traps
    eval $lqphrase ${PRECISION} -o -r ${RANK} $Phrases |
	lqkwic -o "${@-all}" -S '${DocTitle} -- ${DocName}${NL}' -s ''
    set_traps
    short_index_help
}

print_sethelp()
{
    print_info "\
    | @[${SETCMD} match@] [@<precise@>|@<heuristic@>|@<rough@>]
    |	to set the phrase matching level,
    |	currently $PRECISION_DESC
    | @[${SETCMD} index@]
    |	to see a Key Word In Context (KWIC) index of the matches
    | @[${SETCMD} text@]
    |	to browse the actual text that was matched
    | @[${SETCMD} database@] @<directory@>
    |	to use the database in the named directory.
    |	Current value: $LQTEXTDIR
    | @[${SETCMD} rank@] [@<all@>|@<most@>|@<any@>]
    |	@<all@> presents only documents containing all the phrases you typed;
    |	@<most@> shows those first, and then documents with all but one of
    |	the phrases, and so on.
    |	@<any@> doesn${Q}t bother sorting the matches; this is the default,
    |	because it${Q}s the fastest.
    |	Currently: ${RANK_TYPE_DESC}
    | @[${SETCMD} tags@] @<string@>
    |	determine whether SGML tags are shown or hidden in the KWIK index
    | @[${SETCMD} tagchar@] @<string@>
    |	set the character used as a replacement for hidden SGML tags
    | @[${SETCMD} maxhits@] @<n@>|@<all@>
    |	show only the first @<n@> matches (currently $MAXHITS)
    | @[${SETCMD} prompt@] @<string@>
    |	set the prompt for typing in phrases to @<string@>
"
}

set_prompt()
{
    case "$1" in
    default)
	PROMPT="${VBAR} $C"
	;;
    '"'*'"'|"'"*"'")
	PROMPT=`echo "$1" | sed 's/^\(['"${Q}"'""]\)\(.*\)\1/\2/'`
	;;
    ?*)
	PROMPT="$z"
	;;
    *)
	print_info "\
| @[${SETCMD} prompt@] [@<any string@>|@<default@>]
|	sets the prompt to the @<string@> you type, or to
|	the default of ${VBAR} .
|	You can use single or double quotes, as in
|	@<${SETCMD} prompt "enter a phrase: "@>
|	The current prompt is ${PROMPT}.
"
	;;
    esac
    echo $N "${PROMPT}$C"
}

print_set_match_help()
{
    print_info "\
| @[${SETCMD} match@] @<level@>
| Specify the kind of match you want:
| @<precise@>
|       the phrases you type must be matched as
|       exactly as possible in the database.
| @<heuristic@>,@<normal@>,@<default@> (any one will do)
|       CAPITALS, plural@[s@], possessive@[${Q}s@]
|       in your phrases will only match words in the database, e.g.:
|       @<@[Word@]@<   @[Matches@]@<         @>${VBAR}@< @[Word@]@<   @[Matches@]@<                      @>
#       Word   Matches                 |       Word   Matches
|       boys   boys Boys boy${Q}s ${VBAR} boy    boy boys Boy Boys boy${Q}s Boy${Q}s
|       Boys   Boys            ${VBAR} Boy    Boy Boys Boy${Q}s Boys${Q} etc
|       
| @<rough@>,@<approximate@> (either will do)
|       the phrases you type must occur in the original documents
|       but case is ignored, and there can be extra words in the
|       middle of the phrase.
| see @<lqphrase@> @[-xvv@] for more information.
|
"
}

wordgrep()
{
    if test -z "$2"
    then
	print_info "\
    | Use @[${WORDGREP}@] @<pattern@>
    | to see words matching the egrep-style @<pattern@> you give.
"
    else
	# use rs to print the output in columns.
	# rs is included in the lq-text distribution in the src/tools
	# directory.  It's pretty useful...
	lqwordlist -g "$2" |
	$RS
    fi
}

wordlist()
{
    if test -z "$2"
    then
	print_info "\
    | Use @[${WORDGREP}@] @<string@>
    | to see words beginning with the prefix @<string@>.
"
    else
	lqwordlist -s -p "$2" |
	mawk '
	    { printf "%5d %-18s\n", $2, $1 }
	' | $RS
    fi
}

handle_set()
{
    case "$2" in
    *show|text)
	prog=lqshow
	echo "displaying text..."
	;;
    *kwic|*kwik|index)
	prog=lqkwic
	echo "displaying index..."
	;;
    maxhits)
	case "$3" in
	[0-9]|[0-9][0-9]|[0-9][0-9][0-9]|[0-9][0-9]*[0-9][0-9])
	    MAXHITS="$3"
	    ;;
	[nN][oO][nN][eE]|[aA][lL][lL])
	    MAXHITS="all"
	    ;;
	*)
	    print_info "\
| @[${SETCMD} maxhits@] @<number@>|@<all@>
|    Only the first @<n@> lines of output will be shown in the index,
|    unless you explicitly ask for more using a @<range@>, described below.
|    Use @<all@> to see all of the matches by default.
|    The current value is @<$MAXHITS@>.
| 
|    All of the commands such as @[${PAGECMD}@] take a @<range@> of matches
|    as an option.  These can be @[-12@] to specify items up to and including
|    the twelth; @[35-@] to specify items 35 and onward; @[4-7@] to
|    specify items 4, 5, 6 and 7, or @[17@] to specify item 17.  You can
|    separate multiple ranges with spaces or commas, but if you do, the
|    numbers have to increase, so @[-4@] would have to be at the start, and
|    @[12-@] would have to be at the end.
|
|    When you give a @<range@>,
|    the current value of @<maxhits@> is always ignored.
"
	esac
	;;

    database)
	case "$3" in
	?*)
	    if (cd "$3" ) 2> /dev/null
	    then
		echo "${bold}LQTEXTDIR${rmall} set to \"${z}\""
		LQTEXTDIR="$3"; export LQTEXTDIR
	    else
		if test -d "$3"
		then
		    echo "Can${Q}t enter $3 - check permissions"
		else
		    echo "$3 doesn${Q}t seem to be a directory"
		fi
	    fi
	    ;;
	*)
	    print_info "\
| @[${SETCMD} directory@] @<directory@>
| to browse a different ${lqtext} database,
| where @<directory@> is the name of an ${lqtext}
| directory containing @<README@>, @<data@>, etc.
| The default database is:  \$HOME/LQTEXTDIR
| The current value is:     $LQTEXTDIR
|
"
	esac
	;;
    tags)
	case "$3" in
	show*)
	    SGML_CLEVER=;
	    ENTLIST=
	    echo "${VBAR} SGML tags will be shown in ASCII."
	    SGML_TAG_DESC="as ASCII text"
	    ;;
	hid*)
	    SGML_CLEVER="-c -C ${SGML_CHAR}"
	    SGML_TAG_DESC="as ${SGML_CHAR} characters"
	    ENTLIST="-E ${LQTEXTDIR}/entitylist"
	    print_info "\
| Sequences of SGML tags will be shown ${SGML_TAG_DESC}
| for example, the SGML string
|	ele<A><B>phant</B> indeed!
| will become
|	ele${SGML_CHAR}phant${SGML_CHAR} indeed!
| use @[${SETCMD} tagchar@] to change the character used.
"
	    ;;
	*)
	    print_info "\
| @[${SETCMD} tags@] [@<shown@>|@<hidden@>]
|	sets the way SGML tags are presented in the index:
| @<show@> or @<shown@>
|	The tags will be shown as ASCII characters.
| @<hide@> or @<hidden@>
|	The tags will not be shown, but wherever one or more tags was
|	suppressed, a special mark (currently \"${SGML_CHAR}\") will appear.
| 	You can change this character with @[${SETCMD} tagchar{rmbold} @<c@>
|	where @<c@> is the character you want to use.
"
	esac
	;;

    tagchar)
	case "$3" in
	0[0-9]*)
	    SGML_CHAR=`/usr/5bin/echo "\\\\$3"`
	    ;;
	?)
	    SGML_CHAR="$3"
	    ;;
	*)
	    print_info "\
| @[${SETCMD} tagchar@] @<c@>
| Here, @<c@> can be a single character, or it can be an octal (base 8)
| number beginning with a zero, which will be used as a font index for
| a single character.
"
	;;
	esac
	SGML_CLEVER="-c -C ${SGML_CHAR}"
	SGML_TAG_DESC="as ${SGML_CHAR} characters"
	print_info "\
| SGML tag character is now \"${SGML_CHAR}\"
| Sequences of SGML tags will be shown ${SGML_TAG_DESC}
| for example, the SGML string
|	ele<A><B>font</B> you say?
| will become
|	ele${SGML_CHAR}font${SGML_CHAR} you say?
"
	;;
    match)
	case "$3" in
	precise|exact)
	    PRECISION="-mp"
	    PRECISION_DESC="precise matching"
	    ;;
	heuristic|default|normal)
	    PRECISION="-mh"
	    PRECISION_DESC="heuristic matching (the default)"
	    ;;
	rough|approximate)
	    PRECISION="-ma"
	    PRECISION_DESC="approximate (rough) matching"
	    ;;
	*)
	    print_set_match_help
	;;
	esac
	print_info "\
| Currently using $PRECISION_DESC
"
    ;;
    prompt)
	PROMPT=`set_prompt "$3"`
	;;
    rank)
	case "$3" in
	all|and)
	    RANK=all
	    RANK_TYPE_DESC="find documents containing ${smul}all${rmul} of the phrases"
	    ;;
	most)
	    echo "Sorry, ${bold}rank${rmbold} ${smul}most${rmul} unimplemented."
	    ;;
	any|or|none)
	    RANK=none
	    RANK_TYPE_DESC="find documents containing ${smul}any${rmul} of the phrases"
	    ;;
	*)
	    print_info "\
| @[${SETCMD} rank@] [@<all@>|@<most@>|@<any@>]
|	to set the phrase matching level,
| @<all@>
|	Documents will be retrieved if they contain @<all@> of
|	the phrases you type;
| @<most@>
|	Documents will be retrieved if they contain @<any@> of
|	the phrases you type, but the more of the phrases a document contains,
|	the earlier it will appear in the results.
| @<any@>
|	Documents will be retrieved if they contain @<any@> of
|	the phrases you type.
"
	;;
	esac
    ;;
    *) 
	print_sethelp
       ;;
    esac
}

Phrases=
Shown=no
export Phrases Shown

while  [ x"$quit" != x"yes" ]
do
    print_info "\
| Type words or phrases to find, one per line, followed by a blank line.
| Use control-D to quit.  Type @<?@> for more information.
"
    x='fhdjfd'
    y=
    while [ x"$x$y" != x"" ]
    do
	echo $N "${PROMPT}$C"

	read x y z || exit

	case "$x" in
	${SETCMD})
	    handle_set "$x" "$y" "$z"
	    continue
	    ;;
	${VIEWCMD})
	    handle_view "$x" "$y" "$z"
	    continue
	    ;;
	${INDEX})
	    show_index "$x" "$y" "$z"
	    continue
	    ;;
	${LESSCMD})
	    handle_less "$x" "$y" "$z"
	    continue
	    ;;
	${PAGECMD})
	    handle_page "$x" "$y" "$z"
	    continue
	    ;;
	${FILES})
	    show_filenames
	    continue
	    ;;
	${WORDGREP})
	    wordgrep "$x" "$y" "$z"
	    continue
	    ;;
	${WORDLIST})
	    wordlist "$x" "$y" "$z"
	    continue
	    ;;
	${SHELL_CMD})
	    handle_shell "$x" "$y" "$z"
	    continue
	    ;;
	:trace)
	    set -xv
	    continue
	    ;;
	:notrace)
	    set +xv
	    continue
	    ;;
	:*)
	    command_help "$x"
	    continue
	    ;;
	'?')
	    main_help
	    ;;

	?*)
	    if [ x"$Shown" = x"yes" ]
	    then
		Phrases=;
		Shown=no
	    fi

	    # doubling of the quote is for shchk, the shell checker:
	    New=`echo "$x $y $z" | sed 's/"/:/g'`
	    Phrases="${Phrases} \"$New\""
	    ;;
	esac
    done
    if [ x"$prog" = x"lqkwic" ]
    then
	show_index ""
    else
	show_text ""
    fi
    Shown=yes
    # echo $N "Type q to quit, or return to continue: $C"
    # read quit
    # case "$quit" in
    # [qQ]*) quit="yes" ;;
    # *)	   quit=no ;;
    # esac
done
